@@ -38,9 +38,21 @@ import sys
 import traceback
 
 from envstack import __version__, config
-from envstack.env import clear, export, load_environ, resolve_environ, trace_var
+from envstack.env import (
+    bake_environ,
+    Env,
+    encrypt_environ,
+    export_env_to_shell,
+    export,
+    load_environ,
+    resolve_environ,
+    trace_var,
+)
+from envstack.logger import setup_stream_handler
 from envstack.wrapper import run_command
 
+setup_stream_handler()
+
 
 def parse_args():
     """Command line argument parser.
@@ -51,7 +63,7 @@ def parse_args():
 
     if "--" in sys.argv:
         dash_index = sys.argv.index("--")
-        args_after_dash = sys.argv[dash_index + 1 :]
+        args_after_dash = sys.argv[dash_index + 1 :]  # noqa: E203
         args_before_dash = sys.argv[1:dash_index]
     else:
         args_after_dash = []
@@ -73,16 +85,45 @@ def parse_args():
         default=[config.DEFAULT_NAMESPACE],
         help="the environment stacks to use (default '%s')" % config.DEFAULT_NAMESPACE,
     )
-    parser.add_argument(
+    encrypt_group = parser.add_argument_group("encryption options")
+    encrypt_group.add_argument(
+        "-e",
+        "--encrypt",
+        action="store_true",
+        help="encrypt environment values",
+    )
+    encrypt_group.add_argument(
+        "--keygen",
+        action="store_true",
+        help="generate encryption keys",
+    )
+    parser.add_argument_group(encrypt_group)
+    bake_group = parser.add_argument_group("bake options")
+    bake_group.add_argument(
+        "-o",
+        "--out",
+        metavar="FILENAME",
+        help="save the environment to an env file",
+    )
+    bake_group.add_argument(
+        "--depth",
+        type=int,
+        default=0,
+        help="depth of environment stack to bake",
+    )
+    parser.add_argument_group(bake_group)
+    export_group = parser.add_argument_group("export options")
+    export_group.add_argument(
         "--clear",
         action="store_true",
         help="generate unset commands for %s" % config.SHELL,
     )
-    parser.add_argument(
+    export_group.add_argument(
         "--export",
         action="store_true",
         help="generate export commands for %s" % config.SHELL,
     )
+    parser.add_argument_group(export_group)
     parser.add_argument(
         "-p",
         "--platform",
@@ -90,6 +131,13 @@ def parse_args():
         metavar="PLATFORM",
         help="platform to resolve variables for (linux, darwin, windows)",
     )
+    parser.add_argument(
+        "-s",
+        "--set",
+        nargs="*",
+        metavar="VAR:VALUE",
+        help="set a key:value pair in the environment",
+    )
     parser.add_argument(
         "--scope",
         metavar="SCOPE",
@@ -121,9 +169,13 @@ def parse_args():
 
 
 def whichenv():
-    """Entry point for the whichenv command line tool"""
+    """Entry point for the whichenv command line tool. Finds {VAR}s."""
     from envstack.util import findenv
 
+    if len(sys.argv) != 2:
+        print("Usage: whichenv [VAR]")
+        return 2
+
     var_name = sys.argv[1]
     paths = findenv(var_name)
     for path in paths:
@@ -137,30 +189,74 @@ def main():
     try:
         if command:
             return run_command(command, args.namespace)
+
+        elif args.keygen:
+            from envstack.encrypt import generate_keys
+
+            data = generate_keys()
+
+            if args.export:
+                print(export_env_to_shell(data))
+            elif args.out:
+                Env(data).write(args.out)
+            else:
+                for key, value in data.items():
+                    print(f"{key}={value}")
+
+        elif args.set:
+            data = dict(kv.split(":", 1) for kv in args.set)
+
+            if args.encrypt:
+                data = encrypt_environ(data, encrypt=(not args.out))
+            if args.export:
+                print(export_env_to_shell(data))
+            elif args.out:
+                Env(data).write(args.out)
+            else:
+                for key, value in data.items():
+                    print(f"{key}={value}")
+
+        elif args.out:
+            bake_environ(
+                args.namespace,
+                filename=args.out,
+                depth=args.depth or 0,
+                encrypt=args.encrypt,
+            )
+
         elif args.resolve is not None:
             resolved = resolve_environ(
                 load_environ(args.namespace, platform=args.platform)
             )
             keys = args.resolve or resolved.keys()
-            for key in sorted(keys):
+            for key in sorted(str(k) for k in keys):
                 val = resolved.get(key)
                 print(f"{key}={val}")
+
         elif args.trace is not None:
             if len(args.trace) == 0:
                 args.trace = load_environ(args.namespace).keys()
             for trace in args.trace:
                 path = trace_var(*args.namespace, var=trace)
                 print("{0}: {1}".format(trace, path))
+
         elif args.sources:
             env = load_environ(args.namespace, platform=args.platform)
             for source in env.sources:
                 print(source.path)
+
         elif args.clear:
+            from envstack.env import clear
+
             print(clear(args.namespace, config.SHELL))
+
         elif args.export:
             print(export(args.namespace, config.SHELL))
+
         else:
-            env = load_environ(args.namespace, platform=args.platform)
+            env = load_environ(
+                args.namespace, platform=args.platform, encrypt=args.encrypt
+            )
             for k, v in sorted(env.items(), key=lambda x: str(x[0])):
                 print(f"{k}={v}")
 
